use libarchive::read::{Archive, Builder};
use libarchive::{Entry, ReadFilter};
use std::fmt;
use std::io::{self, BufRead, BufReader, Read};
use std::path::{Path, PathBuf};
use std::fs;

const IGNORED_FILES: [&str; 5] = [".BUILDINFO", ".INSTALL", ".MTREE", ".PKGINFO", ".CHANGELOG"];

#[derive(Debug)]
pub struct Package {
    path: PathBuf,
    info: PkgInfo,
    files: Vec<PathBuf>,
    compression: ReadFilter,
}

#[derive(Debug)]
#[derive(Default)]
pub struct PkgInfo {
    name: String,
    base: String,
    version: String,
    description: String,
    size: u64,
    csize: u64,
    url: String,
    arch: String,
    build_date: i64,
    packager: String,
    pgpsig: String,
    pgpsigsize: i64,
    groups: Vec<String>,
    licenses: Vec<String>,
    replaces: Vec<String>,
    depends: Vec<String>,
    conflicts: Vec<String>,
    provides: Vec<String>,
    optdepends: Vec<String>,
    makedepends: Vec<String>,
    checkdepends: Vec<String>,
}

#[derive(Debug, PartialEq, Eq)]
pub enum ParsePkgInfoError {
    InvalidSize,
    InvalidBuildDate,
    InvalidPgpSigSize,
}

impl fmt::Display for ParsePkgInfoError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let s = match self {
            Self::InvalidSize => "invalid size",
            Self::InvalidBuildDate => "invalid build date",
            Self::InvalidPgpSigSize => "invalid pgp sig size",
        };

        write!(f, "{}", s)
    }
}

impl PkgInfo {
    pub fn extend<S: AsRef<str>>(&mut self, line: S) -> Result<(), ParsePkgInfoError> {
        let line = line.as_ref();

        if !line.starts_with('#') {
            if let Some((key, value)) = line.split_once('=').map(|(k, v)| (k.trim(), v.trim())) {
                match key {
                    "pkgname" => self.name = value.to_string(),
                    "pkgbase" => self.base = value.to_string(),
                    "pkgver" => self.version = value.to_string(),
                    "pkgdesc" => self.description = value.to_string(),
                    "size" => {
                        self.size = value.parse().map_err(|_| ParsePkgInfoError::InvalidSize)?
                    }
                    "url" => self.url = value.to_string(),
                    "arch" => self.arch = value.to_string(),
                    "builddate" => {
                        self.build_date = value
                            .parse()
                            .map_err(|_| ParsePkgInfoError::InvalidBuildDate)?
                    }
                    "packager" => self.packager = value.to_string(),
                    "pgpsig" => self.pgpsig = value.to_string(),
                    "pgpsigsize" => {
                        self.pgpsigsize = value
                            .parse()
                            .map_err(|_| ParsePkgInfoError::InvalidPgpSigSize)?
                    }
                    "group" => self.groups.push(value.to_string()),
                    "license" => self.licenses.push(value.to_string()),
                    "replaces" => self.replaces.push(value.to_string()),
                    "depend" => self.depends.push(value.to_string()),
                    "conflict" => self.conflicts.push(value.to_string()),
                    "provides" => self.provides.push(value.to_string()),
                    "optdepend" => self.optdepends.push(value.to_string()),
                    "makedepend" => self.makedepends.push(value.to_string()),
                    "checkdepend" => self.checkdepends.push(value.to_string()),
                    _ => (),
                }
            }
        }

        Ok(())
    }

    pub fn parse<R: Read>(reader: R) -> io::Result<Self> {
        let mut info = Self::default();
        let buf_reader = BufReader::new(reader);

        for line in buf_reader.lines() {
            info.extend(line?).map_err(|e| {
                io::Error::new(io::ErrorKind::Other, format!("pkg info parse error: {}", e))
            })?;
        }

        Ok(info)
    }
}

impl Package {
    pub fn open<P: AsRef<Path>>(path: P) -> io::Result<Self> {
        let mut builder = Builder::new();
        builder.support_format(libarchive::ReadFormat::Tar)?;

        // There are chosen kind of arbitrarily, most likely only zstd, gzip and xz will ever be
        // used
        builder.support_filter(libarchive::ReadFilter::Zstd)?;
        builder.support_filter(libarchive::ReadFilter::Gzip)?;
        builder.support_filter(libarchive::ReadFilter::Xz)?;
        builder.support_filter(libarchive::ReadFilter::Bzip2)?;
        builder.support_filter(libarchive::ReadFilter::Lzma)?;

        let mut ar = builder.open_file(path.as_ref())?;

        let compression = ar.filter(0).ok_or(io::Error::new(io::ErrorKind::Other, "Unknown compression type."))?;
        let mut info: Option<PkgInfo> = None;
        let mut files: Vec<PathBuf> = Vec::new();

        for entry in ar.entries() {
            let entry = entry?;
            let path_name = entry.pathname();

            if !IGNORED_FILES.iter().any(|p| p == &path_name) {
                files.push(PathBuf::from(path_name));
            }

            if path_name == ".PKGINFO" {
                info = Some(PkgInfo::parse(entry)?);
            }
        }

        if let Some(mut info) = info {
            info.csize = fs::metadata(path.as_ref())?.len();

            Ok(Package{
                path: path.as_ref().to_path_buf(),
                info,
                compression,
                files
            })
        } else {
            Err(io::Error::new(io::ErrorKind::Other, "No .PKGINFO file found."))
        }
    }
}
